'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.fireOnSequence = exports.dispatchActionWhen = exports.SEQ = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };
/**
 * Created by andrew on 10/13/16.
 */
/**
 * Schedules actions with { meta: { delay: N } } to be delayed by N milliseconds.
 * Makes `dispatch` return a function to cancel the timeout in this case.
 */


var _fluxStandardAction = require('flux-standard-action');

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _type = require('./type');

var _type2 = _interopRequireDefault(_type);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var localNS = Symbol('SEQ');
var symNext = Symbol('next');
var symRestart = Symbol('restart');
var symRequired = Symbol('required');
var symEmpty = Symbol('empty');
var symTruthy = Symbol('truthy');
var symFalsey = Symbol('falsey');

/**
 * Internal utility function that wraps result function
 * @param fn
 * @param tokenDescription
 * @param replacement
 * @returns {*}
 */
function tagResultFunction(fn, tokenDescription, replacement) {
  fn.tag = localNS;
  if (replacement) {
    tokenDescription = fn.toString().replace(tokenDescription, replacement);
  }
  fn.toString = function () {
    return tokenDescription;
  };
  return fn;
}

function restart(tokens) {
  tokens.forEach(function (t) {
    return t(symRestart);
  });
}

function compare(template, obj) {
  if (_type2.default.isObject(template) && _type2.default.isObject(obj)) {
    return Object.keys(template).reduce(function (m, key) {
      return m && compare(template[key], obj[key]);
    }, true);
  } else {
    switch (template) {
      case symEmpty:
        return _type2.default.isUndefined(obj);
      case symRequired:
        return !_type2.default.isUndefined(obj);
      case symTruthy:
        return !!obj;
      case symFalsey:
        return !obj;
      default:
        return template === obj;
    }
  }
}

var singleSequenceCreator = function singleSequenceCreator(token) {
  var _ref = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var exact = _ref.exact;

  (0, _invariant2.default)(token, 'Token expected for <SINGLE>');

  var tokenStr = null;

  switch (typeof token === 'undefined' ? 'undefined' : _typeof(token)) {
    case 'string':
      tokenStr = token;
      break;

    case 'function':
      if (token.tag === localNS) {
        return token;
      }
      tokenStr = token.toString();
      break;

    case 'object':
      {
        if (exact) {
          return exactObjectSequenceCreator(token);
        } else if (_type2.default.isString(token.type)) {
          tokenStr = token.type;
        }
      }
      break;
  }

  (0, _invariant2.default)(tokenStr, 'Token invalid type for <SINGLE>');

  var result = function result(a) {
    return a !== symRestart && (a.type == tokenStr ? symNext : false);
  };
  var tokenDescription = '<SINGLE>:(' + tokenStr + ')';
  return tagResultFunction(result, tokenDescription);
};

var singleExactSequenceCreator = function singleExactSequenceCreator(token) {
  (0, _invariant2.default)(token && _type2.default.isObject(token), 'Token of plain object type expected for <EXACT>');

  return exactObjectSequenceCreator(token);
};

var exactObjectSequenceCreator = function exactObjectSequenceCreator(token) {
  var result = function result(a) {
    return a !== symRestart && (compare(token, a) ? symNext : false);
  };
  var tokenDescription = '<SINGLE>:(' + JSON.stringify(token) + ')';
  return tagResultFunction(result, tokenDescription);
};

var timesSequenceCreator = function timesSequenceCreator(token, times) {
  var _ref2 = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var strict = _ref2.strict;

  (0, _invariant2.default)(token, 'Token expected for <TIMES>');

  var normalizedToken = singleSequenceCreator(token);
  var tokenDescription = '<TIMES>:(' + normalizedToken.toString() + ' x ' + times + ')';

  var count = 0;
  var result = function result(a) {
    if (a === symRestart) {
      count = 0;
      normalizedToken(symRestart);
      return false;
    }
    var result = normalizedToken(a);
    if (result === true) {
      return true;
    }

    if (result === symNext) {
      count = ++count % times;
      if (count === 0) {
        return symNext;
      } else {
        return true;
      }
    }

    if (strict) {
      count = 0;
      normalizedToken(symRestart);
      return false;
    }
    return true;
  };

  return tagResultFunction(result, tokenDescription);
};

var timesStrictSequenceCreator = function timesStrictSequenceCreator(token, times) {
  (0, _invariant2.default)(token, 'Token expected for <TIMES_STRICT>');

  return tagResultFunction(timesSequenceCreator(token, times, { strict: true }), '<TIMES>', '<TIMES_STRICT>');
};

var allSequenceCreator = function allSequenceCreator(tokens) {
  var _ref3 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref3.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <ALL>');

  var normalizedTokens = tokens.map(singleSequenceCreator);
  var tokenDescription = '<ALL>:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var results = normalizedTokens.map(function (t) {
    return false;
  });
  var length = normalizedTokens.length;

  var result = function result(a) {
    if (a === symRestart) {
      results = normalizedTokens.map(function (t) {
        return false;
      });
      restart(normalizedTokens);
      return false;
    }

    var nextResults = normalizedTokens.map(function (t, k) {
      var result = t(a);

      if (result === symNext) {
        return symNext;
      }
      return results[k];
    }, results);

    var nextResultsReadyCount = nextResults.filter(function (i) {
      return i === symNext;
    }).length;
    if (nextResultsReadyCount === length) {
      results = normalizedTokens.map(function (t) {
        return false;
      });
      restart(normalizedTokens);
      return symNext;
    } else if (strict && nextResultsReadyCount - results.length !== 1) {
      restart(normalizedTokens);
      return false;
    }

    results = nextResults;
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var anySequenceCreator = function anySequenceCreator(tokens) {
  var _ref4 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref4.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <ANY>');

  var normalizedTokens = tokens.map(singleSequenceCreator);
  var tokenDescription = '<ANY>:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var result = function result(a) {
    if (a === symRestart) {
      restart(normalizedTokens);
      return false;
    }
    if (normalizedTokens.filter(function (t) {
      return t(a) === symNext;
    }).length) {
      restart(normalizedTokens);
      return symNext;
    }
    if (strict) {
      restart(normalizedTokens);
      return false;
    }
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var queueSequenceCreator = function queueSequenceCreator(tokens) {
  var _ref5 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref5.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <QUEUE>');

  var normalizedTokens = tokens.map(singleSequenceCreator);
  var tokenDescription = '<QUEUE>:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var length = normalizedTokens.length;
  var count = 0;

  var result = function result(a) {
    if (a === symRestart) {
      count = 0;
      restart(normalizedTokens);
      return false;
    }
    var result = normalizedTokens[count](a);
    if (result === true) {
      return true;
    }

    if (result === symNext) {
      count = ++count % length;

      if (count === 0) {
        restart(normalizedTokens);
        return symNext;
      } else {
        return true;
      }
    }

    if (strict) {
      count = 0;
      restart(normalizedTokens);
      return false;
    }
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var queueStrictSequenceCreator = function queueStrictSequenceCreator(tokens) {
  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <QUEUE_STRICT>');

  return tagResultFunction(queueSequenceCreator(tokens, { strict: true }), '<QUEUE>', '<QUEUE_STRICT>');
};

var sequenceApi = {

  SINGLE: singleSequenceCreator,
  EXACT: singleExactSequenceCreator,
  TIMES: timesSequenceCreator,
  TIMES_STRICT: timesStrictSequenceCreator,
  ALL: allSequenceCreator,
  ANY: anySequenceCreator,
  /**
   * @deprecated
   */
  SEQUENCE: queueSequenceCreator,
  QUEUE: queueSequenceCreator,
  /**
   * @deprecated
   */
  SEQUENCE_STRICT: queueStrictSequenceCreator,
  QUEUE_STRICT: queueStrictSequenceCreator,

  PRESENT: symRequired,
  MISSING: symEmpty,
  TRUTHY: symTruthy,
  FALSEY: symFalsey
};

/**
 * @deprecated: use 'sequenceBuilderCb(s => ...) 'callback in 'dispatchWhen'
 * @type {{SINGLE: singleSequenceCreator, TIMES: timesSequenceCreator, TIMES_STRICT: timesStrictSequenceCreator, ALL: allSequenceCreator, ANY: anySequenceCreator, SEQUENCE: queueSequenceCreator, QUEUE: queueSequenceCreator, SEQUENCE_STRICT: queueStrictSequenceCreator, QUEUE_STRICT: queueStrictSequenceCreator}}
 */
var SEQ = exports.SEQ = sequenceApi;

var sequences = [];

var registerSequence = function registerSequence(fn) {

  sequences.push(fn);

  // unregister
  return function () {
    var idx = sequences.indexOf(fn);
    if (idx >= 0) {
      return sequences.splice(idx, 1);
    }
  };
};

var dispatchActionWhen = exports.dispatchActionWhen = function dispatchActionWhen(action, sequenceBuilderCb) {
  var _ref6 = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var once = _ref6.once;

  return function (dispatch) {
    var sequenceRaw = sequenceBuilderCb(sequenceApi);
    var sequence = null;
    if (_type2.default.isArray(sequenceRaw)) {
      sequence = queueSequenceCreator(sequenceRaw);
      once = true;
    } else if (_type2.default.isObject(sequenceRaw)) {
      sequence = singleSequenceCreator(sequenceRaw);
      once = true;
    } else {
      sequence = singleSequenceCreator(sequenceRaw);
    }

    (0, _invariant2.default)(sequence && _type2.default.isFunction(sequence), 'dispatchWhen expects sequenceBuilderCb to return compiled sequence (function)');

    var seqKey = 'Sequence: ' + sequence.toString() + ' => ' + (action.type || action.toString());

    var unregister = registerSequence(function (a) {

      var result = sequence(a);
      if (result !== symNext) {
        return;
      }

      // console.log(seqKey + ': RESOLVED');

      var unregisterFn = function unregisterFn() {
        unregister && unregister();
        unregister = null;
      };

      if (!once) {
        sequence(symRestart);
      }

      if (_type2.default.isFunction(action)) {
        dispatch(action(unregisterFn));
      } else if (_type2.default.isString(action)) {
        dispatch({
          type: action,
          payload: seqKey,
          meta: {
            sequence: unregisterFn
          }
        });
      } else {
        dispatch(action);
      }

      if (once && unregister) {
        unregisterFn();
        unregisterFn = null;
      }
    });
  };
};

/**
 * @deprecated: use dispatchWhen
 * @param action
 * @param sequence
 * @param once
 * @returns {function(*)}
 */
var fireOnSequence = exports.fireOnSequence = function fireOnSequence(action, sequence, _ref7) {
  var once = _ref7.once;

  return dispatchActionWhen(action, function () {
    return _type2.default.isArray(sequence) ? queueSequenceCreator(sequence) : singleSequenceCreator(sequence);
  }, { once: once });
};

exports.default = function (store) {
  return function (next) {
    return function (action) {
      var result = next(action);

      if (sequences.length && (0, _fluxStandardAction.isFSA)(action)) {
        sequences.forEach(function (s) {
          return s(action);
        });
      }

      return result;
    };
  };
};